package kc.mega.misc;

import kc.mega.BeepBoop;
import kc.mega.dodge.MovementWave;
import kc.mega.game.BotState;
import kc.mega.game.GameState;
import kc.mega.utils.Geom;
import kc.mega.utils.KUtils;
import kc.mega.wave.Bullet;
import kc.mega.wave.Wave;
import kc.mega.wave.WaveManager;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import robocode.AdvancedRobot;
import robocode.BulletHitBulletEvent;
import robocode.HitByBulletEvent;
import robocode.util.Utils;


public class Shielder {
  private static final int POWERS = 10;
  private static final double WIGGLE_SIZE = 0.1;

  private final AdvancedRobot bot;
  private final WaveManager.EnemyWaveManager waveManager;
  private final List<Plan> plans;
  private final List<BulletPredictor> bulletPredictors;

  private int wiggleDirection = 1;
  private int lastOrbitDirection = 1;

  public Shielder(AdvancedRobot bot, WaveManager.EnemyWaveManager waveManager) {
    this.bot = bot;
    this.waveManager = waveManager;
    plans = new ArrayList<>();
    bulletPredictors = new ArrayList<>();
    bulletPredictors.add(new ApproxAbsBearingPredictor());
    bulletPredictors.add(new AbsBearingPredictor());
    bulletPredictors.add(new PrevAbsBearingPredictor());
    bulletPredictors.add(new NoAdjustPrevAbsBearingPredictor());
    bulletPredictors.add(new PrevAbsBearingWithDirectionalOffsetPredictor());
  }

  public void init() {
    if (bot.getRoundNum() > 0) {
      System.out.println("Bullet shielding using " + getBestPredictor().getClass().getSimpleName());
    }
    plans.clear();
  }

  public void shield(GameState gs) {
    if (gs.enemyHistory.size() > 2) {
      BotState state = gs.myHistory.get(2);
      if (state.velocity != 0) {
        lastOrbitDirection = KUtils.orbitDirection(gs.enemyHistory.get(2).absoluteBearing(state),
            state.heading, (int)Math.signum(state.velocity));
      }
    }
    if (gs.enemyFiredLastTick || gs.gameTime == 20) {
      MovementWave w = (MovementWave)new MovementWave.Builder()
          .isVirtual(false)
          .myHistory(gs.enemyHistory.subList(1, gs.enemyHistory.size()))
          .enemyHistory(gs.myHistory.subList(1, gs.myHistory.size()))
          .power(gs.lastEnemyBulletPower)
          .hasBullet(gs.enemyFiredLastTick)
          .doPaint(BeepBoop.paint)
          .lastOrbitDirection(lastOrbitDirection)
          .build();
      if (gs.gameTime != 20) {  // time 20 "warmup" seems to help with not skipping turn first wave
        waveManager.add(w);
      }

      wiggleDirection *= -1;
      List<Plan> options = new ArrayList<>();
      long fireTime = gs.gameTime + Math.max(1,
          (long)Math.ceil(bot.getGunHeat() / bot.getGunCoolingRate()));
      double predictedBulletHeading = getBestPredictor().predictBulletHeading(w);
      for (boolean wiggle : new boolean[] {true, false}) {
        for (double bulletPower : getPowers(gs, w, fireTime, wiggle, predictedBulletHeading)) {
          options.add(new Plan(gs, w, bulletPower, fireTime, wiggle, predictedBulletHeading));
        }
      }
      if (options.size() > 0 && gs.gameTime != 20) {
        plans.add(options.stream().max((p1, p2) -> (int)Math.signum(p1.score - p2.score)).get());
      }
    }
    for (MovementWave w : waveManager.updateWaves(gs.getMyState())) {
      gs.enemyHitRateTracker.logShotPassed(w.power);
    }

    if (plans.isEmpty()) {
      bot.setAhead(0);
      bot.setTurnRightRadians(Utils.normalRelativeAngle(parallelHeading(
          gs.getMyState().absoluteBearing(gs.getEnemyState())) - bot.getHeadingRadians()));
      bot.setTurnGunRightRadians(Utils.normalRelativeAngle(
          gs.getMyState().absoluteBearing(gs.getEnemyState()) - bot.getGunHeadingRadians()));
    } else if (plans.get(0).execute(gs.gameTime)) {
      plans.remove(0);
    }
  }

  private double parallelHeading(double absBearing) {
    return Math.sin(bot.getHeadingRadians() - absBearing) > 0 ?
        absBearing + Math.PI/2 : absBearing - Math.PI/2;
  }

  public List<Double> getPowers(GameState gs, MovementWave targetWave, long fireTime,
      boolean wiggle, double predictedBulletHeading) {
    List<Double> powers = new ArrayList<>();
    for (int i = 0; i < POWERS; i++) {
      powers.add(0.1 * Math.pow(Math.pow(Math.max(0.2, targetWave.power) * 10, 1.0 / POWERS), i));
    }

    double heading = parallelHeading(Geom.absoluteBearingPrecise(
        gs.getMyState().location, targetWave.source));
    Point2D.Double fireLocation = wiggle ? Geom.projectPrecise(
        gs.getMyState().location, heading, wiggleDirection * WIGGLE_SIZE) :  gs.getMyState().location;
    Bullet enemyBullet = new Bullet(
        targetWave.source, predictedBulletHeading,
        targetWave.speed, targetWave.fireTime);
    Point2D.Double bulletPosition = enemyBullet.getMidpoint(fireTime);
    double bulletDist = bulletPosition.distance(fireLocation);
    double lastBulletDist = Double.POSITIVE_INFINITY;
    long time = fireTime;
    while (bulletDist < lastBulletDist) {
      lastBulletDist = bulletDist;
      time++;
      bulletPosition = enemyBullet.getMidpoint(time);
      bulletDist = bulletPosition.distance(fireLocation);
      double neededSpeed = bulletDist / (time - fireTime + 0.5);
      if (neededSpeed > 11 && neededSpeed < 19.7) {
        double power = (20 - neededSpeed) / 3;
        if (power < targetWave.power) {
          powers.add(power);
        }
      }
    }
    return powers;
  }

  private class Plan {
    private final long fireTime;
    private final double bulletPower;
    private final double fireAngle;
    private final double score;
    private final double heading;
    private final boolean wiggle;
    private final int direction;

    public Plan(GameState gs, MovementWave targetWave, double bulletPower, long fireTime,
        boolean wiggle, double predictedBulletHeading) {
      this.fireTime = fireTime;
      this.bulletPower = bulletPower;
      this.wiggle = wiggle;
      this.direction = wiggleDirection;

      heading = parallelHeading(Geom.absoluteBearingPrecise(
          gs.getMyState().location, targetWave.source));
      Point2D.Double fireLocation = wiggle ? Geom.projectPrecise(
          gs.getMyState().location, heading, direction * WIGGLE_SIZE) : gs.getMyState().location;
      Wave myWave = new Wave.Builder()
          .trackPreciseShadows(true)
          .myHistory(Collections.singletonList(new BotState(
              fireLocation, 0, 0, 0, 0, 0, 0, fireTime)))
          .enemyHistory(gs.enemyFuture.subList(
              Math.min(gs.enemyFuture.size() - 2, (int)(fireTime - gs.gameTime)),
              gs.enemyFuture.size()))
          .power(bulletPower)
          .build();
      Bullet enemyBullet = new Bullet(
          targetWave.source, predictedBulletHeading,
          targetWave.speed, targetWave.fireTime);
      myWave.addBulletShadows(enemyBullet, fireTime);

      Wave.Shadow bestShadow = null;
      double bestShadowScore = 0;
      double width = 0;
      for (Wave.Shadow shadow : myWave.preciseShadows) {
        double score = shadow.weight + (shadow.endOffset - shadow.startOffset) / 100;
        if (score > bestShadowScore) {
          bestShadowScore = score;
          bestShadow = shadow;
          width = shadow.endOffset - shadow.startOffset;
        }
      }
      fireAngle = myWave.absoluteBearing + (bestShadow == null ? 0 :
        bestShadow.endOffset + bestShadow.startOffset) / 2;
      double targetPower = targetWave.power * gs.getMyState().energy /
          (gs.getEnemyState().energy + targetWave.power);
      double targetPowerConservative = 0.9 * targetWave.power *
          Math.max(0.1, gs.getMyState().energy - 1) /
          (gs.getEnemyState().energy + targetWave.power - 1);

      score = bestShadowScore
          + (width > 1e-6 ? 0.2 : 0)
          + (width > 1e-5 ? 0.03 : 0)
          + (!wiggle && width > 1e-4 ? 0.1 : 0)
          + (bulletPower < targetPower ? 0.01 : 0)
          + (bulletPower < targetPowerConservative ? 0.01 : 0)
          + width
          - bulletPower * 1e-10;
    }

    public boolean execute(long gameTime) {
      bot.setTurnRightRadians(Utils.normalRelativeAngle(heading - bot.getHeadingRadians()));
      bot.setTurnGunRightRadians(Utils.normalRelativeAngle(fireAngle - bot.getGunHeadingRadians()));
      if (gameTime == fireTime - 1 && wiggle) {
        bot.setAhead(direction * WIGGLE_SIZE);
      } else if (gameTime == fireTime) {
        bot.setFire(bulletPower);
        if (wiggle) {
          bot.setAhead(-direction * WIGGLE_SIZE);
        }
      }
      return gameTime >= fireTime;
    }
  }

  private BulletPredictor getBestPredictor() {
    return bulletPredictors.stream().max((a1, a2) -> (int)Math.signum(a1.score - a2.score)).get();
  }

  private class PrevAbsBearingPredictor extends BulletPredictor {
    @Override
    public double getHeadOnBearing(MovementWave w) {return w.prevAbsoluteBearing;}
    @Override
    public double getStartingScore() {return 5e-5;}
  }

  private class PrevAbsBearingWithDirectionalOffsetPredictor extends PrevAbsBearingPredictor {
    @Override
    public double adjustOffset(double offset, MovementWave w) {return offset * w.orbitDirection;}
    @Override
    public double getStartingScore() {return 1e-5;}
  }

  private class NoAdjustPrevAbsBearingPredictor extends BulletPredictor {
    @Override
    public double getHeadOnBearing(MovementWave w) {
      return w.prevAbsoluteBearing + w.fireState.heading - w.myState.heading;
    }
    @Override
    public double getStartingScore() {return 4e-5;}
  }

  private class ApproxAbsBearingPredictor extends BulletPredictor {
    @Override
    public double getHeadOnBearing(MovementWave w) {
      return Geom.absoluteBearingPrecise(w.myState.getNextState(0).location, w.enemyState.location);
    }
    @Override
    public double getStartingScore() {return 3e-5;}
  }

  private class AbsBearingPredictor extends BulletPredictor {
    @Override
    public double getHeadOnBearing(MovementWave w) {return w.absoluteBearing;}
    @Override
    public double getStartingScore() {return 2e-5;}
  }

  private static abstract class BulletPredictor {
    public final List<Double> hitOffsets;
    public double mostLikelyOffset;
    public double score;

    public BulletPredictor() {
      hitOffsets = new ArrayList<>();
      setOffsetAndScore();
    }

    public abstract double getStartingScore();

    public abstract double getHeadOnBearing(MovementWave w);

    public double predictBulletHeading(MovementWave w) {
      return getHeadOnBearing(w) + adjustOffset(mostLikelyOffset, w);
    }

    public void logBullet(MovementWave w, double bulletHeading) {
      hitOffsets.add(adjustOffset(Utils.normalRelativeAngle(
          bulletHeading - getHeadOnBearing(w)), w));
      if (hitOffsets.size() > 500) {
        hitOffsets.remove(0);
      }
      setOffsetAndScore();
    }

    public double adjustOffset(double offset, MovementWave w) {
      return offset;
    }

    public double getOffsetScore(int offsetIndex, int historySize) {
      double offset = hitOffsets.get(offsetIndex);
      double offsetScore = (Math.abs(offset) < 0.01 ? 2 : 0) - Math.abs(offset) +
          offsetIndex * 1e-8;
      for (int j = 0; j < Math.min(hitOffsets.size(), historySize); j++) {
        if (offsetIndex != j && Math.abs(hitOffsets.get(j) - offset) < 1e-5) {
          offsetScore += 1;
        }
      }
      return offsetScore;
    }

    public void setOffsetAndScore() {
      double bestScore = 0;
      int bestIndex = 0;
      for (int i = 0; i < hitOffsets.size(); i++) {
        double offsetScore = getOffsetScore(i, 20);
        if (offsetScore > bestScore) {
          bestScore = offsetScore;
          bestIndex = i;
        }
      }
      mostLikelyOffset = hitOffsets.size() == 0 ? 0 : hitOffsets.get(bestIndex);
      score = getStartingScore() + (hitOffsets.size() == 0 ? 0 : getOffsetScore(bestIndex, 1000));
    }
  }

  public void onHitByBullet(HitByBulletEvent e, GameState gs) {
    MovementWave hitWave = waveManager.getHitWave(e.getPower());
    if (hitWave != null && hitWave.fireTime < gs.gameTime - 1) {
      logBullet(hitWave, e.getHeadingRadians());
    }
  }

  public void onBulletHitBullet(BulletHitBulletEvent e) {
    MovementWave hitWave = waveManager.getBulletHitBulletWave(
        new Point2D.Double(e.getHitBullet().getX(), e.getHitBullet().getY()),
        e.getHitBullet().getPower(), bot.getTime());
    if (hitWave != null) {
      logBullet(hitWave, e.getHitBullet().getHeadingRadians());
    }
  }

  public void logBullet(MovementWave wave, double bulletHeading) {
    for (BulletPredictor bulletPredictor : bulletPredictors) {
      bulletPredictor.logBullet(wave, bulletHeading);
    }
  }
}
